/**
    @file kbd.c

    @brief This will emulate a keyboard and print out the message specified 
 *         in the string tty_message on the host. The message will be written 
 *         in a teletype style.

    @version 1.0.0

**/
/*
 * ============================================================================
 * History
 * =======
 *
 * (C) Copyright 2014, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */
 
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "ft51.h"
#include "ft51_general_config.h"
#include "ft51_interrupts.h"
#include "ft51_usb.h"
#include "ft51_usb_dfu.h"
#include "ft51_usb_hid.h"
#include "ft51_dfu.h"

/** @brief String to 'type' on the host.
 *
 *  @details The string is translated in key scan codes, effectively looking like a
 *   key press to the host. The ASCII characters in the string are translated into
 *   scan codes using a look-up table.
 *   There is the capability of sending control keys (e.g. page up, function keys,
 *   delete, insert, caps lock etc) using an arbitrary escape character. In this app
 *   the character is a backslash. The next letter after this is decoded from the
 *   control scan code table rather than the ASCII scan code table. So, "\\C" will
 *   send a caps lock. This escape character can be changed in the main function.
 *   \r or \x0d is carriage return.
 *   \x08 is backspace
 *   \x1b is escape
 *   \x09 is horizontal tab
 *   \\C, \\S, \\N are caps lock, scroll lock and num lock
 *   \\r, \\l, \\d, \\u are right, left, up, down arrows
 *   \\H, \\E, \\U, \\D are home, end, page up and page down
 *   \\X is delete
 *   \\I is insert
 */
__code char tty_message[] =
        "Hello, I am an \\Cftdi ft51a\\C device... nice to meet you!\r";

/**
 @name USB and Hub Configuration
 @brief Indication of whether the USB Hub is enabled, how it is
        powered and the size of the control endpoint Max Packets.
 */
//@{
// USB Hub Enabled - set to 1 for enabled and 0 for disabled
#ifndef USB_HUB_ENABLED
#define USB_HUB_ENABLED 1
#endif // USB_HUB_ENABLED
// USB Bus Powered - set to 1 for self-powered or 0 for bus-powered
#ifndef USB_SELF_POWERED
#define USB_SELF_POWERED 0
#endif // USB_SELF_POWERED
#if USB_SELF_POWERED == 1
#define USB_CONFIG_BMATTRIBUTES_VALUE (USB_CONFIG_BMATTRIBUTES_SELF_POWERED | USB_CONFIG_BMATTRIBUTES_RESERVED_SET_TO_1)
#else // USB_SELF_POWERED
#define USB_CONFIG_BMATTRIBUTES_VALUE USB_CONFIG_BMATTRIBUTES_RESERVED_SET_TO_1
#endif // USB_SELF_POWERED
// USB Endpoint Zero packet size (both must match)
#define USB_CONTROL_EP_MAX_PACKET_SIZE 8
#define USB_CONTROL_EP_SIZE USB_EP_SIZE_8
//@}

/**
 @name Device Configuration Areas
 @brief Size and location of string descriptors and
 structure used to modify device and configuration
 descriptors.
 */
//@{
// String descriptors - allow a maximum of 256 bytes for this
#define STRING_DESCRIPTOR_LOCATION 0x80
#define STRING_DESCRIPTOR_ALLOCATION 0x100
//@}

/**
 @name DFU_BLOCK_SIZE definition.
 @brief Number of bytes in block, sent in each DFU_DNLOAD request
 from the DFU update program on the host. This is simplified
 in that the meaning of a block is an arbitrary number of
 bytes. This is intentionally a multiple of the maximum
 packet size for the control endpoints.
 */
//@{
#define DFU_BLOCK_SIZE 16
//@}

/**
 @name USB_PID_KEYBOARD configuration.
 @brief Run Time Product ID for keyboard function.
 */
//@{
#define USB_PID_KEYBOARD 0x0fea
//@}

/**
 @name DFU_USB_PID_DFUMODE configuration.
 @brief FTDI predefined DFU Mode Product ID.
 */
//@{
#define DFU_USB_PID_DFUMODE 0x0fee
//@}

/**
 @name DFU_USB_INTERFACE configuration..
 @brief Run Time and DFU Mode Interface Numbers.
 */
//@{
#define DFU_USB_INTERFACE_RUNTIME 1
#define DFU_USB_INTERFACE_DFUMODE 0
//@}

/**
 @name WCID_VENDOR_REQUEST_CODE for WCID.
 @brief Unique vendor request code for WCID OS Vendor Extension validation.
*/
//@{
#define WCID_VENDOR_REQUEST_CODE	 0xF1
//@}

/**
 @def MICROSECONDS_TO_TIMER_TICKS
 @brief Macro to calculate the number of timer ticks required to
 create a 1ms counter.
 Scale is clock frequency in MHz (assume 48) / clock-cycles per instruction.
 16-bit timer counts up to 65536, so subtract duration from 65536.
 */
#define MICROSECONDS_TO_TIMER_TICKS(a) (65536 - ((a) * (48 / 12)))

/**
 @name int_in
 @brief Interrupt IN endpoint setup status.
 */
static FT51_STATUS int_in = FT51_FAILED;

/**
 @name report_protocol
 @brief Flag to set report protocol (0 - boot, 1 - report)
 */
uint8_t report_protocol = 0;

/**
 @name report_enable
 @brief Flag to enable report mode (i.e. reporting active).
 */
static uint8_t report_enable = 0;

/**
 @name report_idle
 @brief Timer to set idle report period.
 */
static uint8_t report_idle = 0;

/**
 @name input_report
 @brief Flag to enable input reports (i.e. LEDs).
 */
static uint8_t input_report = 0;

/**
 @name keypress_counter
 @brief Count up ms timer to simulate keypress reports at a set interval.
 */
static volatile int16_t keypress_counter = 0;

/**
 @name string_descriptor
 @brief Table of USB String descriptors

 This is placed at a fixed location in the __code section allowing
 up-to 256 bytes of string descriptors to be defined. These can be
 modified or replaced by a utility or binary editor without
 requiring a recompilation of the firmware.
 They are placed at offset 0x100 and reserve 0x100 bytes.
 The data is not stored in __xdata section and is therefore
 regarded as const.
 */
__code __at(STRING_DESCRIPTOR_LOCATION)uint8_t string_descriptor[STRING_DESCRIPTOR_ALLOCATION] =
{
    // String 0 is actually a list of language IDs supported by the real strings.
    0x04, USB_DESCRIPTOR_TYPE_STRING, 0x09,
    0x04, // 0409 = English (US)

    // String 1 (Manufacturer): "FTDI"
    0x0a, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, 'D', 0x00, 'I',
    0x00,

    // String 2 (Product): "FT51A Keyboard"
    0x1E, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, '5', 0x00, '1',
    0x00, 'A', 0x00, ' ', 0x00, 'K', 0x00, 'e', 0x00, 'y', 0x00, 'b', 0x00,
    'o', 0x00, 'a', 0x00, 'r', 0x00, 'd', 0x00,

    // String 3 (Serial Number): "FT424242"
    0x12, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, '4', 0x00, '2',
    0x00, '4', 0x00, '2', 0x00, '4', 0x00, '2', 0x00,

    // String 4 (DFU Product Name): "FT51A DFU Mode"
    0x1E, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, '5', 0x00, '1',
    0x00, 'A', 0x00, ' ', 0x00, 'D', 0x00, 'F', 0x00, 'U', 0x00, ' ', 0x00,
    'M', 0x00, 'o', 0x00, 'd', 0x00, 'e', 0x00,

    // String 5 (Interface Name): "DFU Interface"
    0x1c, USB_DESCRIPTOR_TYPE_STRING, 'D', 0x00, 'F', 0x00, 'U', 0x00, ' ',
    0x00, 'I', 0x00, 'n', 0x00, 't', 0x00, 'e', 0x00, 'r', 0x00, 'f', 0x00,
    'a', 0x00, 'c', 0x00, 'e', 0x00,

    // END OF STRINGS
    0x00
};

/**
 @name wcid_string_descriptor
 @brief USB String descriptor for WCID identification.
 */
__code uint8_t wcid_string_descriptor[USB_MICROSOFT_WCID_STRING_LENGTH] = {
		USB_MICROSOFT_WCID_STRING(WCID_VENDOR_REQUEST_CODE)
};

/**
 @name device_descriptor_keyboard
 @brief Device descriptor for Run Time mode.
 */
__code USB_device_descriptor device_descriptor_keyboard =
{
    .bLength = 0x12,
    .bDescriptorType = USB_DESCRIPTOR_TYPE_DEVICE,
    .bcdUSB = USB_BCD_VERSION_2_0,          // V2.0
    .bDeviceClass = USB_CLASS_DEVICE,       // Defined in interface
    .bDeviceSubClass = USB_SUBCLASS_DEVICE, // Defined in interface
    .bDeviceProtocol = USB_PROTOCOL_DEVICE, // Defined in interface
    .bMaxPacketSize0 = USB_CONTROL_EP_MAX_PACKET_SIZE,
    .idVendor = USB_VID_FTDI,   // idVendor: 0x0403 (FTDI)
    .idProduct = USB_PID_KEYBOARD, // idProduct: 0x0fea
    .bcdDevice = 0x0101,        // 1.1
    .iManufacturer = 0x01,      // Manufacturer
    .iProduct = 0x02,           // Product
    .iSerialNumber = 0x03,      // Serial Number
    .bNumConfigurations = 0x01
};

/**
 @name config_descriptor_keyboard
 @brief Config descriptor for Run Time mode.
 */
//@{
__code struct config_descriptor_keyboard
{
    USB_configuration_descriptor configuration;
    USB_interface_descriptor interface;
    USB_hid_descriptor hid;
    USB_endpoint_descriptor endpoint;
    USB_interface_descriptor dfu_interface;
    USB_dfu_functional_descriptor dfu_functional;
};

// Fields marked with * are updated with information in device_config[]
struct config_descriptor_keyboard config_descriptor_keyboard =
{
    .configuration.bLength = 0x09,
    .configuration.bDescriptorType = USB_DESCRIPTOR_TYPE_CONFIGURATION,
    .configuration.wTotalLength = sizeof(struct config_descriptor_keyboard),
    .configuration.bNumInterfaces = 0x02,
    .configuration.bConfigurationValue = 0x01,
    .configuration.iConfiguration = 0x00,
    .configuration.bmAttributes =  USB_CONFIG_BMATTRIBUTES_VALUE,
    .configuration.bMaxPower = 0xFA,           // 500mA

	// ---- INTERFACE DESCRIPTOR for Keyboard ----
    .interface.bLength = 0x09,
    .interface.bDescriptorType = USB_DESCRIPTOR_TYPE_INTERFACE,
    .interface.bInterfaceNumber = 0,
    .interface.bAlternateSetting = 0x00,
    .interface.bNumEndpoints = 0x01,
    .interface.bInterfaceClass = USB_CLASS_HID, // HID Class
    .interface.bInterfaceSubClass = USB_SUBCLASS_HID_BOOT_INTERFACE, // Boot Protocol
    .interface.bInterfaceProtocol = USB_PROTOCOL_HID_KEYBOARD, // Keyboard
    .interface.iInterface = 0x02,               // "FT51A Keyboard"

    // ---- HID DESCRIPTOR for Keyboard ----
    .hid.bLength = 0x09,
    .hid.bDescriptorType = USB_DESCRIPTOR_TYPE_HID,
    .hid.bcdHID = USB_BCD_VERSION_HID_1_1,
    .hid.bCountryCode = USB_HID_LANG_NOT_SUPPORTED,
    .hid.bNumDescriptors = 0x01,
    .hid.bDescriptorType_0 = USB_DESCRIPTOR_TYPE_REPORT,
    .hid.wDescriptorLength_0 = 0x0041,

    // ---- ENDPOINT DESCRIPTOR for Keyboard ----
    .endpoint.bLength = 0x07,
    .endpoint.bDescriptorType = USB_DESCRIPTOR_TYPE_ENDPOINT,
    .endpoint.bEndpointAddress = 0x81,
    .endpoint.bmAttributes = USB_ENDPOINT_DESCRIPTOR_ATTR_INTERRUPT,
    .endpoint.wMaxPacketSize = 0x0008,
    .endpoint.bInterval = 0x0a,

    // ---- INTERFACE DESCRIPTOR for DFU Interface ----
    .dfu_interface.bLength = 0x09,
    .dfu_interface.bDescriptorType = USB_DESCRIPTOR_TYPE_INTERFACE,
    .dfu_interface.bInterfaceNumber = DFU_USB_INTERFACE_RUNTIME,
    .dfu_interface.bAlternateSetting = 0x00,
    .dfu_interface.bNumEndpoints = 0x00,
    .dfu_interface.bInterfaceClass = USB_CLASS_APPLICATION, // Application Specific Class
    .dfu_interface.bInterfaceSubClass = USB_SUBCLASS_DFU, // Device Firmware Update
    .dfu_interface.bInterfaceProtocol = USB_PROTOCOL_DFU_RUNTIME, // Runtime Protocol
    .dfu_interface.iInterface = 0x05,       // "DFU Interface"

    // ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
    .dfu_functional.bLength = 0x09,
    .dfu_functional.bDescriptorType = USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL,
    .dfu_functional.bmAttributes = USB_DFU_BMATTRIBUTES_CANDNLOAD | USB_DFU_BMATTRIBUTES_WILLDETACH, // bitCanDnload and bitWillDetach
    .dfu_functional.wDetatchTimeOut = DFU_TIMEOUT, // suggest 8192ms
    .dfu_functional.wTransferSize = DFU_BLOCK_SIZE, // make it 16 bytes
    .dfu_functional.bcdDfuVersion = USB_BCD_VERSION_DFU_1_1, // DFU Version 1.1
};
//@}

/**
 @name wcid_feature_runtime
 @brief WCID Compatible ID for DFU interface in runtime.
 */
//@{
__code USB_WCID_feature_descriptor wcid_feature_runtime =
{
	sizeof(struct USB_WCID_feature_descriptor), /* dwLength */
	USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
	USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID, /* wIndex */
	1, /* bCount */
	{0, 0, 0, 0, 0, 0, 0,}, /* rsv1 */
	DFU_USB_INTERFACE_RUNTIME, /* bFirstInterfaceNumber */
	1, /* rsv2 - set to 1 */
	{'W', 'I', 'N', 'U', 'S', 'B', 0x00, 0x00,}, /* compatibleID[8] */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,}, /* subCompatibleID[8] */
	{0, 0, 0, 0, 0, 0,}, /* rsv3[6] */
};
//@}

/**
 @name wcid_feature_dfumode
 @brief WCID Compatible ID for DFU interface in DFU mode.
 */
//@{
__code USB_WCID_feature_descriptor wcid_feature_dfumode =
{
	sizeof(struct USB_WCID_feature_descriptor), /* dwLength */
	USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
	USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID, /* wIndex */
	1, /* bCount */
	{0, 0, 0, 0, 0, 0, 0,}, /* rsv1 */
	DFU_USB_INTERFACE_DFUMODE, /* bFirstInterfaceNumber */
	1, /* rsv2 - set to 1 */
	{'W', 'I', 'N', 'U', 'S', 'B', 0x00, 0x00,}, /* compatibleID[8] */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,}, /* subCompatibleID[8] */
	{0, 0, 0, 0, 0, 0,}, /* rsv3[6] */
};
//@}

/**
 @name wcid_device_guid_dfu
 @brief WCID Compatible DeviceInterfaceGUIDs for DFU interface.
 */
//@{
__code USB_WCID_extended_property_descriptor wcid_device_guid_dfu =
{
	sizeof(struct USB_WCID_extended_property_descriptor), /* dwLength */
	USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
	USB_MICROSOFT_WCID_FEATURE_WINDEX_DEVICE_GUID, /* wIndex */
	1, /* wCount */
	136, /* dwSize */
	USB_MICROSOFT_WCID_PROPERTY_DATATYPE_STRING_MULTI, /* dwPropertyDataType */
	42, /* wPropertyNameLength */
	{'D', 0x00, 'e', 0x00, 'v', 0x00, 'i', 0x00, 'c', 0x00, 'e', 0x00,
	 'I', 0x00, 'n', 0x00, 't', 0x00, 'e', 0x00, 'r', 0x00, 'f', 0x00,
	 'a', 0x00, 'c', 0x00, 'e', 0x00, 'G', 0x00, 'U', 0x00, 'I', 0x00,
	 'D', 0x00, 's', 0x00, 0x00, 0x00 }, /* null terminated PropertyName */
	80, /* dwPropertyDataLength */
	/* GUID for FT51A DFU interfaces "{1eaaaa95-1bd6-49c2-b4ae-286583f61227}" */
	{'{', 0x00, '1', 0x00, 'e', 0x00, 'a', 0x00, 'a', 0x00, 'a', 0x00,
	 'a', 0x00, '9', 0x00, '5', 0x00, '-', 0x00, '1', 0x00, 'b', 0x00,
	 'd', 0x00, '6', 0x00, '-', 0x00, '4', 0x00, '9', 0x00, 'c', 0x00,
	 '2', 0x00, '-', 0x00, 'b', 0x00, '4', 0x00, 'a', 0x00, 'e', 0x00,
	 '-', 0x00, '2', 0x00, '8', 0x00, '6', 0x00, '5', 0x00, '8', 0x00,
	 '3', 0x00, 'f', 0x00, '6', 0x00, '1', 0x00, '2', 0x00, '2', 0x00,
	 '7', 0x00, '}', 0x00,
	 0x00, 0x00, 0x00, 0x00}, /* bPropertyData */
};
//@}

/**
 @name hid_report_descriptor_keyboard
 @brief HID Report descriptor for keyboard.

 See Device Class Definition for Human Interface Devices (HID) Version 1.11
 from USB Implementers Forum USB.org

 0x05, 0x01,             Usage Page: Generic Desktop Controls
 0x09, 0x06,             Usage: Keyboard
 0xA1, 0x01,             Collection: Application
 0x05, 0x07,               Usage Page: Keyboard
 0x19, 0xE0,               Usage Minimum: Keyboard LeftControl
 0x29, 0xE7,               Usage Maximum: Keyboard Right GUI
 0x15, 0x00,               Logical Minimum: 0
 0x25, 0x01,               Logical Maximum: 1
 0x75, 0x01,               Report Size: 1
 0x95, 0x08,               Report Count: 8
 0x81, 0x02,               Input: Data (2)
 0x95, 0x01,               Report Count: 1
 0x75, 0x08,               Report Size: 8
 0x81, 0x01,               Input: Constant (1)
 0x95, 0x03,               Report Count: 3
 0x75, 0x01,               Report Size: 1
 0x05, 0x08,               Usage Page: LEDs
 0x19, 0x01,               Usage Minimum: Num Lock
 0x29, 0x03,               Usage Maximum: Scroll Lock
 0x91, 0x02,               Output: Data (2)
 0x95, 0x05,               Report Count: 5
 0x75, 0x01,               Report Size: 1
 0x91, 0x01,               Output: Constant (1)
 0x95, 0x06,               Report Count: 6
 0x75, 0x08,               Report Size: 8
 0x15, 0x00,               Logical Minimum: 0
 0x26, 0xFF, 0x00,         Logical Maximum: 255
 0x05, 0x07,               Usage Page: Keyboard/Keypad
 0x19, 0x00,               Usage Minimum: 0
 0x2A, 0xFF, 0x00,         Usage Maximum: 255
 0x81, 0x00,               Input: Data (0)
 0xC0                    End collection
 **/
//@{
__code uint8_t hid_report_descriptor_keyboard[65] =
{ 0x05, 0x01, 0x09, 0x06, 0xA1, 0x01, 0x05, 0x07, 0x19, 0xE0, 0x29, 0xE7, 0x15,
        0x00, 0x25, 0x01, 0x75, 0x01, 0x95, 0x08, 0x81, 0x02, 0x95, 0x01, 0x75,
        0x08, 0x81, 0x01, 0x95, 0x03, 0x75, 0x01, 0x05, 0x08, 0x19, 0x01, 0x29,
        0x03, 0x91, 0x02, 0x95, 0x05, 0x75, 0x01, 0x91, 0x01, 0x95, 0x06, 0x75,
        0x08, 0x15, 0x00, 0x26, 0xFF, 0x00, 0x05, 0x07, 0x19, 0x00, 0x2A, 0xFF,
        0x00, 0x81, 0x00, 0xC0, };
//@}

/**
 @struct hid_report_structure
 @brief HID Report structure to match
 hid_report_descriptor_keyboard descriptor.
 **/
typedef struct hid_report_structure_t
{
    unsigned char kbdLeftControl :1;
    unsigned char kbdLeftShift :1;
    unsigned char kbdLeftAlt :1;
    unsigned char kbdLeftGUI :1;
    unsigned char kbdRightControl :1;
    unsigned char kbdRightShift :1;
    unsigned char kbdRightAlt :1;
    unsigned char kbdRightGUI :1;

    unsigned char arrayNotUsed;  // [1]
    unsigned char arrayKeyboard; // [2]
    unsigned char arrayResv1;
    unsigned char arrayResv2;
    unsigned char arrayResv3;
    unsigned char arrayResv4;
    unsigned char arrayResv5;
} hid_report_structure_t;

/**
 @name report_buffer
 @brief Buffer used to report back to host the current state of keypresses.
 */
hid_report_structure_t report_buffer;

/**
 @name device_descriptor_dfumode
 @brief Device descriptor for DFU Mode.
 */
__code USB_device_descriptor device_descriptor_dfumode =
{
    .bLength = 0x12,
    .bDescriptorType = USB_DESCRIPTOR_TYPE_DEVICE,
    .bcdUSB = USB_BCD_VERSION_2_0,          // V2.00
    .bDeviceClass = USB_CLASS_DEVICE,       // Defined in interface
    .bDeviceSubClass = USB_SUBCLASS_DEVICE, // Defined in interface
    .bDeviceProtocol = USB_PROTOCOL_DEVICE, // Defined in interface
    .bMaxPacketSize0 = USB_CONTROL_EP_MAX_PACKET_SIZE,    // 8
    .idVendor = USB_VID_FTDI,   // idVendor: 0x0403 (FTDI)
    .idProduct = DFU_USB_PID_DFUMODE, // idProduct: 0x0fee
    .bcdDevice = 0x0101,        // 1.1
    .iManufacturer = 0x01,      // Manufacturer
    .iProduct = 0x04,           // Product
    .iSerialNumber = 0x03,      // Serial Number
    .bNumConfigurations = 0x01
};

/**
 @name config_descriptor_dfumode
 @brief Config descriptor for DFU Mode.
 */
//@{
__code struct config_descriptor_dfumode
{
    USB_configuration_descriptor configuration;
    USB_interface_descriptor dfu_interface;
    USB_dfu_functional_descriptor dfu_functional;
};

// Fields marked with * are updated with information in device_config[]
struct config_descriptor_dfumode config_descriptor_dfumode =
{
    .configuration.bLength = 0x09,
    .configuration.bDescriptorType = USB_DESCRIPTOR_TYPE_CONFIGURATION,
    .configuration.wTotalLength = sizeof(struct config_descriptor_dfumode),
    .configuration.bNumInterfaces = 0x01,
    .configuration.bConfigurationValue = 0x01,
    .configuration.iConfiguration = 0x00,
    .configuration.bmAttributes = USB_CONFIG_BMATTRIBUTES_VALUE,
    .configuration.bMaxPower = 0xFA, // 500mA

    // ---- INTERFACE DESCRIPTOR for DFU Interface ----
    .dfu_interface.bLength = 0x09,
    .dfu_interface.bDescriptorType = USB_DESCRIPTOR_TYPE_INTERFACE,
    .dfu_interface.bInterfaceNumber = DFU_USB_INTERFACE_DFUMODE,
    .dfu_interface.bAlternateSetting = 0x00,
    .dfu_interface.bNumEndpoints = 0x00,
    .dfu_interface.bInterfaceClass = USB_CLASS_APPLICATION, // Application Specific Class
    .dfu_interface.bInterfaceSubClass = USB_SUBCLASS_DFU, // Device Firmware Update
    .dfu_interface.bInterfaceProtocol = USB_PROTOCOL_DFU_DFUMODE, // Runtime Protocol
    .dfu_interface.iInterface = 0x05,
    // ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
    .dfu_functional.bLength = 0x09,
    .dfu_functional.bDescriptorType = USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL,
    .dfu_functional.bmAttributes = USB_DFU_BMATTRIBUTES_CANDNLOAD,  	// bitCanDnload
    .dfu_functional.wDetatchTimeOut = DFU_TIMEOUT, // suggest 8192ms
    .dfu_functional.wTransferSize = DFU_BLOCK_SIZE,     // 16
    .dfu_functional.bcdDfuVersion = USB_BCD_VERSION_DFU_1_1, // DFU Version 1.1
};
//@}

/**
 @name Keyboard Scan Codes Lookup table (for ASCII keys)
 @brief Lookup table for converting ASCII to scan codes.
 */
__code char ASCII_to_scancode[][3] =
{
    { 4, 'a', 'A' },
    { 5, 'b', 'B' },
    { 6, 'c', 'C' },
    { 7, 'd', 'D' },
    { 8, 'e', 'E' },
    { 9, 'f', 'F' },
    { 10, 'g', 'G' },
    { 11, 'h', 'H' },
    { 12, 'i', 'I' },
    { 13, 'j', 'J' },
    { 14, 'k', 'K' },
    { 15, 'l', 'L' },
    { 16, 'm', 'M' },
    { 17, 'n', 'N' },
    { 18, 'o', 'O' },
    { 19, 'p', 'P' },
    { 20, 'q', 'Q' },
    { 21, 'r', 'R' },
    { 22, 's', 'S' },
    { 23, 't', 'T' },
    { 24, 'u', 'U' },
    { 25, 'v', 'V' },
    { 26, 'w', 'W' },
    { 27, 'x', 'X' },
    { 28, 'y', 'Y' },
    { 29, 'z', 'Z' },
    { 30, '1', '!' },
    { 31, '2', '"' },
    { 32, '3', '' },
    { 33, '4', '$' },
    { 34, '5', '%' },
    { 35, '6', '^' },
    { 36, '7', '&' },
    { 37, '8', '*' },
    { 38, '9', '(' },
    { 39, '0', ')' },
    { 40, 0x0D, 0 }, // <CR> Carriage Return (Enter)
    { 41, 0x1B, 0 }, // <ESC> Escape
    { 42, 0x08, 0 }, // <BS> Backspace
    { 43, 0x09, 0 }, // <HT> Horizontal Tab
    { 44, ' ', 0 },
    { 45, '-', '_' },
    { 46, '=', '+' },
    { 47, '[', '{' },
    { 48, ']', '}' },
    { 50, '#', '~' },
    { 51, ';', ':' },
    { 52, '\'', '@' },
    { 54, ',', '<' },
    { 55, '.', '>' },
    { 56, '/', '?' },
    { 100, '\\', '|' },
    { 0, 0, 0 }, // end marker
};

/**
 @name Keyboard Scan Codes Lookup table (for control keys)
 @brief Lookup table for converting control keys to scan codes.
 */
__code char CONTROL_to_scancode[][2] =
{
    { 'r', 0x4D }, // Right
    { 'l', 0x4B }, // Left
    { 'd', 0x50 }, // Down
    { 'u', 0x48 }, // Up
    { 'C', 0x3A }, // Caps Lock
    { 'S', 0x46 }, // Scroll Lock
    { 'N', 0x53 }, // Num Lock
    { '1', 0x3B }, // F1
    { '2', 0x3C }, // F2
    { '3', 0x3D }, // F3
    { '4', 0x3E }, // F4
    { '5', 0x3F }, // F5
    { '6', 0x40 }, // F6
    { '7', 0x41 }, // F7
    { '8', 0x42 }, // F8
    { '9', 0x43 }, // F9
    { 'a', 0x44 }, // F10
    { 'b', 0x57 }, // F11
    { 'c', 0x58 }, // F12
    { 'X', 0x53 }, // DEL
    { 'I', 0x52 }, // INS
    { 'H', 0x47 }, // Home
    { 'U', 0x49 }, // Page Up
    { 'D', 0x51 }, // Page Down
    { 'E', 0x4F }, // End
    { 0, 0 },
};

/**
     @name Countdown timer
         A millisecond counter is used to detach the device when
         required by the DFU.
 */
static volatile uint16_t ms_timer = 0;

// Forward declarations.
void usb_detach_attach(void);
void usb_setup(void);

/**
 @brief      Invoked by ISR for Timer0 every millisecond
 @details    Each millisecond Timer0 interrupt handler will call
 this routine.
 When the detach counter is active (> zero) then we will decrement it.
 If we are in the DFUSTATE_APPDETACH state when it reaches zero then
 we will assume that the host has not send the USB reset required to
 put the device into DFU Mode and revert the state machine back to
 DFUSTATE_APPIDLE to wait for another DFU_DETACH request.
 We also make a timer for the keypress simulator.
 @param[in]  flags (not used).
 **/
void ms_timer_interrupt(const uint8_t flags)
{
    (void) flags; // Flags not currently used

    keypress_counter--;

    // The DFU detach timer must be called once per millisecond
    dfu_timer();

    if (ms_timer > 0)
    {
        ms_timer--;
    }

    // Reload the timer
    TH0 = MSB(MICROSECONDS_TO_TIMER_TICKS(1000));
    TL0 = LSB(MICROSECONDS_TO_TIMER_TICKS(1000));
}

/**
 @brief      Initialise Timer0 to fire every millisecond.
 @details    Associate the ms_timer_interrupt handler function with
 the Timer0 interrupt.
 **/
void ms_timer_initialise(void)
{
    // Register our own handler for interrupts from Timer 0
    interrupts_register(ms_timer_interrupt, interrupts_timer0);

    // Timer0 is controlled by TMOD bits 0 to 3, and TCON bits 4 to 5.
    TMOD &= 0xF0; // Clear Timer0 bits
    TMOD |= 0x01; // Put Timer0 in mode 1 (16 bit)
    // Set the count-up value so that it rolls over to 0 after 1 millisecond.
    TH0 = MSB(MICROSECONDS_TO_TIMER_TICKS(1000));
    TL0 = LSB(MICROSECONDS_TO_TIMER_TICKS(1000));
    TCON &= 0xCF; // Clear Timer0's Overflow and Run flags
    TCON |= 0x10; // Start Timer0 (set its Run flag)
}

/**
 @brief      USB class request handler
 @details    Handle class requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and the DFU state machine and status updated as
 required.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bRequest is not
 supported.
 **/
FT51_STATUS class_req_cb(USB_device_request *req)
{
    FT51_STATUS status = FT51_FAILED;
    uint8_t interface = LSB(req->wIndex) & 0x0F;

    // For DFU requests the SETUP packet must be addressed to the
    // the DFU interface on the device.

    // For DFU requests ensure the recipient is an interface...
    if ((req->bmRequestType & USB_BMREQUESTTYPE_RECIPIENT_MASK) ==
    USB_BMREQUESTTYPE_RECIPIENT_INTERFACE)
    {
        // ...and that the interface is the correct Runtime interface
        if (dfu_is_runtime())
        {
            if ((interface == DFU_USB_INTERFACE_RUNTIME))
            {
                // Handle only DFU_DETATCH, DFU_GETSTATE and DFU_GETSTATUS
                // when in Runtime mode. Table 3.2 DFU_DETACH is mandatory
                // in Runtime mode, DFU_GETSTATE and DFU_GETSTATUS are
                // optional.
                switch (req->bRequest)
                {
                case USB_CLASS_REQUEST_DETACH:
                    dfu_class_req_detach(req->wValue);
                    usb_detach_attach();
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_GETSTATUS:
                    dfu_class_req_getstatus();
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_GETSTATE:
                    dfu_class_req_getstate();
                    status = FT51_OK;
                    break;
                }
            }
            else
            {
                // Handle HID class requests
                switch (req->bRequest)
                {
                case USB_HID_REQUEST_CODE_SET_IDLE:
                    report_enable = 1;
                    report_idle = req->wValue >> 8;
                    USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
                    status = FT51_OK;
                    break;

                case USB_HID_REQUEST_CODE_GET_IDLE:
                    USB_transfer(USB_EP_0, USB_DIR_IN, &report_idle, 1);
                    status = FT51_OK;
                    break;

                case USB_HID_REQUEST_CODE_SET_PROTOCOL:
                    USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
                    report_protocol = req->wValue & 0xff;
                    status = FT51_OK;
                    break;

                case USB_HID_REQUEST_CODE_GET_PROTOCOL:
                    USB_transfer(USB_EP_0, USB_DIR_IN, &report_protocol, 1);
                    status = FT51_OK;
                    break;

                case USB_HID_REQUEST_CODE_GET_REPORT:
					USB_transfer(1, USB_DIR_IN, (char *) &report_buffer, sizeof(report_buffer));
                    status = FT51_OK;
                    break;

                case USB_HID_REQUEST_CODE_SET_REPORT:
                    // dummy read of one byte
                    USB_transfer(USB_EP_0, USB_DIR_OUT, &input_report, 1);
                    // Update LED status on GPIO P0
                    P0 = input_report;
                    // Acknowledge SETUP
                    USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
                    status = FT51_OK;
                    break;
                }
            }
        }
        // ...or the correct DFU Mode interface
        else
        {
            if (interface == DFU_USB_INTERFACE_DFUMODE)
            {
                // Handle remaining DFU class requests when in DFU Mode.
                // Table 3.2 DFU_DETACH is not supported in DFU Mode.
                switch (req->bRequest)
                {
                case USB_CLASS_REQUEST_DNLOAD:
                    /* For simplicity only program the maximum block size
                     * starting at an address which is an exact multiple
                     * of this. The block size is given in the configuration
                     * descriptor for the DFU interface.
                     */
                    dfu_class_req_download(req->wValue * DFU_BLOCK_SIZE,
                            req->wLength);
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_GETSTATUS:
                    dfu_class_req_getstatus();
                    if (dfu_is_wait_reset())
                    {
                    	usb_detach_attach();
                    }
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_GETSTATE:
                    dfu_class_req_getstate();
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_CLRSTATUS:
                    dfu_class_req_clrstatus();
                    status = FT51_OK;
                    break;
                case USB_CLASS_REQUEST_ABORT:
                    dfu_class_req_abort();
                    status = FT51_OK;
                    break;

                default:
                case USB_CLASS_REQUEST_UPLOAD:
                    // Unknown or unsupported request.
                    break;
                }
            }
        }
    }

    return status;
}

/**
 @brief      USB standard request handler for GET_DESCRIPTOR
 @details    Handle standard GET_DESCRIPTOR requests from the host
 application.
 The hValue is parsed and the appropriate device,
 configuration or string descriptor is selected.
 For device and configuration descriptors the DFU
 state machine determines which descriptor to use.
 For string descriptors the lValue selects which
 string from the string_descriptors table to use.
 Strings are obtained from the __code section and
 typecast to a be compatible with __xdata storage.
 The string table is automatically traversed to find
 the correct string and the length is taken from the
 string descriptor.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bmRequestType is not
 supported.
 **/
FT51_STATUS standard_req_get_descriptor(USB_device_request *req)
{
    uint8_t *src = NULL;
    uint16_t length = req->wLength;
    uint8_t hValue = req->wValue >> 8;
    uint8_t lValue = req->wValue & 0x00ff;
    uint8_t i;
    uint8_t slen;

    switch (hValue)
    {
    case USB_DESCRIPTOR_TYPE_DEVICE:

        if (dfu_is_runtime())
        {
            src = (char *) &device_descriptor_keyboard;
        }
        else
        {
            src = (char *) &device_descriptor_dfumode;
        }
        if (length > sizeof(USB_device_descriptor)) // too many bytes requested
            length = sizeof(USB_device_descriptor); // Entire structure.
        break;

    case USB_DESCRIPTOR_TYPE_CONFIGURATION:

        if (dfu_is_runtime())
        {
            src = (char *) &config_descriptor_keyboard;
            if (length > sizeof(config_descriptor_keyboard)) // too many bytes requested
                length = sizeof(config_descriptor_keyboard); // Entire structure.
        }
        else
        {
            src = (char *) &config_descriptor_dfumode;
            if (length > sizeof(config_descriptor_dfumode)) // too many bytes requested
                length = sizeof(config_descriptor_dfumode); // Entire structure.
        }
        break;

    case USB_DESCRIPTOR_TYPE_STRING:

		// Special case: WCID descriptor
		if (lValue == USB_MICROSOFT_WCID_STRING_DESCRIPTOR)
		{
			src = (uint8_t *)wcid_string_descriptor;
			length = sizeof(wcid_string_descriptor);
			break;
		}

        // Find the nth string in the string descriptor table
        i = 0;
        while ((slen = string_descriptor[i]) > 0)
        {
            // Point to start of string descriptor in __code segment
            // Typecast prevents warning for losing const when usb_transfer
            // is called
            src = (uint8_t *) &string_descriptor[i];
            if (lValue == 0)
            {
                break;
            }
            i += slen;
            lValue--;
        }
        if (lValue > 0)
        {
            // String not found
            return FT51_FAILED;
        }
        // Update the length returned only if it is less than the requested
        // size
        if (length > slen)
        {
            length = slen;
        }

        break;

    case USB_DESCRIPTOR_TYPE_REPORT:

        src = (char *) &hid_report_descriptor_keyboard;
        if (length > sizeof(hid_report_descriptor_keyboard))
        {
            length = sizeof(hid_report_descriptor_keyboard); // Entire structure.
        }
        break;

    default:
        return FT51_FAILED;
    }

    USB_transfer(USB_EP_0, USB_DIR_IN, src, length);

    return FT51_OK;
}

/**
 @brief      USB standard request handler for GET_STATUS
 @details    Handle standard GET_STATUS requests from the host
             application.
             The bmRequestType value is parsed and the appropriate
             action or function is performed.
 @param[in]	 req - USB_device_request structure containing the
             SETUP portion of the request from the host.
 @return	 status - FT51_OK if successful or FT51_FAILED
             if there is an error or the bmRequestType is not
             supported.
 **/
FT51_STATUS standard_req_get_status(USB_device_request *req)
{
    USB_STATE state = USB_get_state();
    USB_ENDPOINT_NUMBER ep_number = LSB(req->wIndex) & 0x0F;
    USB_ENDPOINT_DIR ep_dir = (LSB(req->wIndex) >> 7);
    uint8_t buf[2];

    buf[0] = buf[1] = 0;

    // Get Status for endpoints only
    if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_DEV_TO_HOST |
            USB_BMREQUESTTYPE_RECIPIENT_ENDPOINT))
    {
        if (((state < CONFIGURED) && (ep_number == 0))
                || (state >= CONFIGURED))
        {
            if (USB_get_ep_stalled(ep_number, ep_dir))
            {
                buf[0] = USB_GET_STATUS_ENDPOINT_HALT;
            }
            USB_transfer(USB_EP_0, USB_DIR_IN, buf, 2);

            return FT51_OK;
        }
    }
    else if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_DEV_TO_HOST |
            USB_BMREQUESTTYPE_RECIPIENT_DEVICE))
    {
        // This must match the configuration descriptor's bmAttributes
        buf[0] = USB_CONFIG_BMATTRIBUTES_VALUE;
        USB_transfer(USB_EP_0, USB_DIR_IN, buf, 2);

        return FT51_OK;
    }
    return FT51_FAILED;
}

/**
 @brief      USB standard request handler for GET_FEATURE and SET_FEATURE
 @details    Handle standard GET_FEATURE and SET_FEATURE requests
             from the host application.
             The bmRequestType value is parsed and the appropriate
             action or function is performed.
 @param[in]  req - USB_device_request structure containing the
             SETUP portion of the request from the host.
 @return        status - FT51_OK if successful or FT51_FAILED
             if there is an error or the bmRequestType is not
             supported.
 **/
FT51_STATUS standard_req_set_clear_feature(USB_device_request *req)
{
    USB_ENDPOINT_NUMBER ep_number = LSB(req->wIndex) & 0x0F;
    USB_ENDPOINT_DIR ep_dir = (LSB(req->wIndex) >> 7);
    USB_STATE state = USB_get_state();

    if (req->bmRequestType == (USB_BMREQUESTTYPE_DIR_HOST_TO_DEV |
            USB_BMREQUESTTYPE_RECIPIENT_ENDPOINT))
    {
        // Only support the endpoint halt feature in this device
        if (req->wValue == USB_FEATURE_ENDPOINT_HALT)
        {
            // Only allow this if the device is configured or the endpoint
            // is zero if the device is not configured.
            if (((state < CONFIGURED) && (ep_number == 0))
                    || (state >= CONFIGURED))
            {
                // Perform a stall or a clear
                if (req->bRequest == USB_REQUEST_CODE_CLEAR_FEATURE)
                {
                    // Feature selector CLEAR
                    USB_clear_endpoint(ep_number, ep_dir);
                }
                else
                {
                    // Feature selector SET
                    USB_stall_endpoint(ep_number, ep_dir);
                }
                USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
                return FT51_OK;
            }
        }
    }
    return FT51_FAILED;
}

/**
 @brief      USB standard request handler
 @details    Handle standard requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and provided to other handlers.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bRequest is not
 supported.
 **/
FT51_STATUS standard_req_cb(USB_device_request *req)
{
    FT51_STATUS status = FT51_FAILED;

    switch (req->bRequest)
    {
    case USB_REQUEST_CODE_GET_STATUS:
        status = standard_req_get_status(req);
        break;

    case USB_REQUEST_CODE_CLEAR_FEATURE:
    case USB_REQUEST_CODE_SET_FEATURE:
        status = standard_req_set_clear_feature(req);
        break;

    case USB_REQUEST_CODE_GET_DESCRIPTOR:
        status = standard_req_get_descriptor(req);
        break;

    case USB_REQUEST_CODE_SET_CONFIGURATION:
        status = USB_req_set_configuration(req);
        break;

    case USB_REQUEST_CODE_GET_CONFIGURATION:
        status = USB_req_get_configuration();
        break;

    case USB_REQUEST_CODE_SET_ADDRESS:
        status =  USB_req_set_address(req);
        break;

    default:
    case USB_REQUEST_CODE_GET_INTERFACE:
    case USB_REQUEST_CODE_SET_INTERFACE:
        // Unknown or unsupported request.
        status = FT51_FAILED;
        break;
    }

    return status;
}

/**
 @brief      USB vendor request handler
 @details    Handle vendor requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and provided to other handlers.
 There are no vendor requests requiring handling in
 this firmware.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - FT51_OK if successful or FT51_FAILED
 if there is an error or the bRequest is not
 supported.
 **/
FT51_STATUS vendor_req_cb(USB_device_request *req)
{
    FT51_STATUS status = FT51_FAILED;
	uint16_t length = req->wLength;

	// For Microsoft WCID only.
	// Request for Compatible ID Feature Descriptors.
	// Check the request code and wIndex.
	if (req->bRequest == WCID_VENDOR_REQUEST_CODE)
	{
		if (req->bmRequestType & USB_BMREQUESTTYPE_DIR_DEV_TO_HOST)
		{
			switch (req->wIndex)
			{
			case USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID:
				if (length > sizeof(wcid_feature_runtime)) // too many bytes requested
					length = sizeof(wcid_feature_runtime); // Entire structure.
				// Return a compatible ID feature descriptor.
				if (dfu_is_runtime())
				{
					USB_transfer(USB_EP_0, USB_DIR_IN,
						(uint8_t *) &wcid_feature_runtime, length);
					status = FT51_OK;
				}
				else
				{
					USB_transfer(USB_EP_0, USB_DIR_IN,
						(uint8_t *) &wcid_feature_dfumode, length);
					status = FT51_OK;
				}
				break;

			case USB_MICROSOFT_WCID_FEATURE_WINDEX_DEVICE_GUID:
				if (length > sizeof(wcid_device_guid_dfu)) // too many bytes requested
					length = sizeof(wcid_device_guid_dfu); // Entire structure.
				// Return a compatible ID feature descriptor.
				if (((dfu_is_runtime()) && (req->wValue == DFU_USB_INTERFACE_RUNTIME))
					|| ((!dfu_is_runtime()) && (req->wValue == DFU_USB_INTERFACE_DFUMODE)))
				{
					USB_transfer(USB_EP_0, USB_DIR_IN,
							(uint8_t *) &wcid_device_guid_dfu, length);
					status = FT51_OK;
				}
				break;
			}
			// Make sure we terminate SETUP packet with a ZLP for this type of transfer.
			//if ((length % USB_CONTROL_EP_MAX_PACKET_SIZE) == 0)
				//USB_transfer(USB_EP_0, USB_DIR_IN, NULL, 0);
		}
	}
    return status;
}

/**
 @brief      USB suspend callback
 @details    Called when the USB bus enters the suspend state.
 @param[in]	status - unused.
 @return		N/A
 **/
void suspend_cb(uint8_t status)
{
    (void) status;
    return;
}

/**
 @brief      USB resume callback
 @details    Called when the USB bus enters the resume state
 prior to restating after a suspend.
 @param[in]  status - unused.
 @return     N/S
 **/
void resume_cb(uint8_t status)
{
    (void) status;
    return;
}

/**
 @brief      USB reset callback
 @details    Called when the USB bus enters the reset state.
 @param[in]	status - unused.
 @return		N/A
 **/
void reset_cb(uint8_t status)
{
    (void) status;

    USB_free_endpoint(1, USB_DIR_IN);

    USB_set_state(DEFAULT);

    if (dfu_reset())
    {
        // Run downloaded application - reload MTP to shadow RAM
        // Execution effectively stops now and returns to the main()
        // function of the new program in MTP.
        device_revert();
    }

    /* Make application specific BULK, INTERRUPT or
     * ISOCHRONOUS endpoints.
     */
    int_in = USB_create_endpoint(USB_EP_1, USB_EP_INT,
            USB_DIR_IN, USB_EP_SIZE_8);

    return;
}

/**
 @brief      USB detach attach sequence
 @details    Called when the USB device wishes to detach from the
 	 	 	 bus then reattach.
 @return		N/A
 **/
void usb_detach_attach(void)
{
	ms_timer = 10;
	while (ms_timer > 0);
	// Disable USB device to force an reset
	USB_finalise();
	ms_timer = 10;
	while (ms_timer > 0);
	// Start the USB device.
	usb_setup();
}

/**
 @brief      Initialise emulation of a USB device
 @details    Sets up callback functions and data structures
 required to implement control endpoints on device.
 @param[in]	ep_size - Endpoint size of control endpoints.
 @return     N/A
 **/
void usb_setup(void)
{
    USB_ctx usb_ctx;

    memset(&usb_ctx, 0, sizeof(usb_ctx));

    usb_ctx.standard_req = standard_req_cb;
    usb_ctx.class_req = class_req_cb;
    usb_ctx.vendor_req = vendor_req_cb;
    usb_ctx.suspend = suspend_cb;
    usb_ctx.resume = resume_cb;
    usb_ctx.reset = reset_cb;
    usb_ctx.lpm = NULL;

    usb_ctx.hub_en.mask.hub_enable = USB_HUB_ENABLED;
    usb_ctx.hub_en.mask.hub_LocalPwrSrc = USB_SELF_POWERED;
    usb_ctx.hub_en.mask.hub_RmtWkUpEn = 1;

    // Initialise the USB device with a control endpoint size
    // of 8 bytes. This must match the size for bMaxPacketSize0
    // defined in the device descriptor.
    usb_ctx.ep_size = USB_CONTROL_EP_SIZE;

    // Perform USB Device initialisation.
    USB_initialise(&usb_ctx);

    return;
}

// key states
#define KEY_ON      1
#define KEY_OFF     2

// Time between SetIdle and sending keypresses
#define KEY_STARTUP_TIME 2000 // 2 seconds
// How long between key presses (in ms)
#define KEY_OFF_PERIOD 50
#define KEY_OFF_PERIOD_LONG 300
// How long to press key for (in ms)
#define KEY_ON_PERIOD    10

/**
 @brief      main function.
 @details    Modify any configuration parameters which are updatable
 and call initialisation routines.
 Finally process USB requests as they arrive (for ever).
 **/
void main(void)
{
    unsigned char state = KEY_ON; // First action is a keypress.
    unsigned short delay = KEY_ON_PERIOD;
    unsigned char key = 0xFF;
    unsigned char shift;
    __code char *pmsg = tty_message; // Pointer to message to send to host
    // Initialise FT51A
    device_initialise();

    // Initialise USB function
    usb_setup();

    memset(&report_buffer, 0, sizeof(report_buffer));

    // Initialise the timer to provide a millisecond countdown timer
    ms_timer_initialise();

    // Wait an initial delay for OS to setup keyboard.
    keypress_counter = KEY_STARTUP_TIME;

    // Endless loop.
    while (1)
    {
        // Check control endpoints and handle SETUPs
        USB_process();

        /* Keyboard emulation code. */
        /* Note that report_idle timer is not implemented - assumed to be zero (infinite) */
        if ((USB_get_state() == CONFIGURED) && (report_enable))
        {
            if ((key > 0) || (state == KEY_OFF))
            {
                if (keypress_counter < 0)
                {
                    keypress_counter = delay;

                    // Wait until any previous keypress has been sent
                    if (!USB_ep_buffer_full(1, USB_DIR_IN))
                    {
                        switch (state)
                        {
                        case KEY_ON:

                            delay = KEY_ON_PERIOD;
                            state = KEY_OFF;
                            shift = 0;

                            // Check escape code for NON-ASCII keys
                            // This is an arbitrary character to signify an escape code
                            if (*pmsg != '\\')
                            {
                                // ASCII key decode
                                for (key = 0; ASCII_to_scancode[key][0]; key++)
                                {
                                    if (ASCII_to_scancode[key][1] == *pmsg)
                                    {
                                        break;
                                    }
                                    if (ASCII_to_scancode[key][2] == *pmsg)
                                    {
                                        shift = 1;
                                        break;
                                    }
                                }
                                key = ASCII_to_scancode[key][0];
                                pmsg++;
                            }
                            else
                            {
                                // CONTROL key decode
                                // next character is control key
                                pmsg++;
                                for (key = 0; CONTROL_to_scancode[key][0];
                                        key++)
                                {
                                    if (CONTROL_to_scancode[key][1] == *pmsg)
                                    {
                                        break;
                                    }
                                }
                                key = CONTROL_to_scancode[key][0];
                                pmsg++;
                            }

                            report_buffer.arrayKeyboard = key;
                            report_buffer.kbdRightShift = shift;
                            break;

                        case KEY_OFF:

                            delay = KEY_OFF_PERIOD;
                            if (*pmsg == ' ')
                                delay = KEY_OFF_PERIOD_LONG;
                            state = KEY_ON;

                            report_buffer.arrayKeyboard = 0x00;
                            report_buffer.kbdRightShift = 0;

                            // key is non-zero to continue... end of string will stop us.
                            key = *pmsg;
                            break;

                        default:
                            break;
                        }

                        USB_transfer(1, USB_DIR_IN, (char *) &report_buffer, 8);
                    }
                }
            }
        }
    }
}
